/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Rotate the unit square around its corner at the origin.
   <pre>{@code
            y                             y

            |                             |
            |                            /|\
          1 +---------+                / 1+  \
            |         |              /    |    \
            |         |             <     |     >
            |         |              \    |    /
            |         |                \  |  /
            |         |                  \|/
    --------+---------+---> x     --------+--------+-------> x
            |         1                   |        1
            |                             |
            |                             |
   }</pre>
<p>
   See the animation <a href="../Example_01.gif">Example_01.gif</a>.
<p>
   <a href="../Example_01.gif"><img src="../Example_01.gif" alt="Example_01.gif"></a>
<p>
   For the i'th frame of the animation, the square's model matrix is
   <pre>{@code
      M = Rz(5*i)
        = Rz(5*(i-1)) * Rz(5)    // acumulate a rotation on the right
        = Rz(5) * Rz(5*(i-1)).   // acumulate a rotation on the left
   }</pre>
   Notice how each new rotation can be accumulated on the right or left
   end of the matrix product.
<p>
   In the code below, the main loop for creating the animation
   frames looks like this.
   <pre>{@code
      for (int i = 0; i < 72; ++i)
      {
         square_p.transform( Matrix.rotateZ(5 * i) );

         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_01a_Frame%03d.ppm", i));
      }
   }</pre>
   Each frame of the animation rotates the square an additional 5 degrees
   from the previous frame.
<p>
   A second, alternative, loop for creating the animation looks like this.
   <pre>{@code
      square_p.transform( Matrix.identity() );
      for (int i = 0; i < 72; ++i)
      {
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_01b_Frame%03d.ppm", i));

         square_p.transform( square_p.getMatrix().times(Matrix.rotateZ(5)) );
      }
   }</pre>
<p>
   Notice that this code "accumulates" the rotations in the square's model
   matrix. Here is what we mean by that.
<p>
   Let {@code M} denote the model matrix for the square. Initially,
   {@code M} is the identity matrix.
   <pre>{@code
               M = [ 1  0  0  0 ]
                   [ 0  1  0  0 ]
                   [ 0  0  1  0 ]
                   [ 0  0  0  1 ]
   }</pre>
<p>
   Let {@code Rz(a)} denote the 4x4 homogeneous matrix that rotates the
   xy-plane around the z-axis by angle {@code a} degrees counter-clockwise.
   <pre>{@code
           Rz(a) = [ cos(a * (pi/180))   -sin(a * (pi/180))   0   0 ]
                   [ sin(a * (pi/180))    cos(a * (pi/180))   0   0 ]
                   [       0                    0             1   0 ]
                   [       0                    0             0   1 ]
   }</pre>
<p>
   Let {@code square_p} refer to the {@code Position} object that refers
   to our {@code Model} of the square and its model {@code Matrix}.
<p>
   Each time we execute the line of code
   <pre>{@code
       square_p.transform( square_p.getMatrix().times(Matrix.rotateZ(5)) );
   }</pre>
   we multiply the position's matrix {@code M} on the RIGHT by the matrix
   {@code Rz(5)} and the new matrix value replaces the current value in
   {@code square_p}. Essentially, this line of code looks like this.
   <pre>{@code
          M = M * Rz(5).
   }</pre>
   Before the loop begins, {@code M} is the identity matrix,
   <pre>{@code
          M = I.
   }</pre>
   After one iteration of the loop {@code M} is
   <pre>{@code
          M = I * Rz(5).
   }</pre>
   After two iterations, {@code M} is
   <pre>{@code
          M = I * Rz(5) * Rz(5)
   }</pre>
   but this simplifies to
   <pre>{@code
          M = Rz(10).
   }</pre>
   After {@code i} iterations, {@code M} simplifies to
   <pre>{@code
          M = Rz(5*i).
   }</pre>
   So the Position's matrix is just accumulating the total rotation
   and each new frame accumulates a new rotation on the right hand
   side of the previous frame's tranformation.
<p>
   Here is the scene tree that this program builds for the i'th frame
   of the animation. There is only one Matrix object in the scene. All
   of the transformation information needs to be in that single matrix.
   Notice how the additional transformation needed by each new frame, a
   rotation by 5 degrees, is accumulated in the Position's matrix.
   <pre>{@code
             Scene
            /     \
           /       \
     Camera         List<Position>
                          |
                          |
                      Position
                     /        \
                    /          \
              Matrix            Square
              Rz(5i)
   }</pre>
*/
public class Example_01
{
   public static void main(String[] args)
   {
      // Since this is a 2D example, we can use an orthographic camera.
      final double right  = 2.0;
      final double left   = -right;
      final double top    = right;
      final double bottom = -top;
      final Scene scene = new Scene("Example_01",
                                 Camera.projOrtho(left, right, bottom, top));

      // Create a FrameBuffer to render our scene into.
      final int width  = 512;
      final int height = 512;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.black);

      // Create a set of x and y axes.
      final Model axes = new Axes2D(left, right, bottom, top, 8, 8);
      ModelShading.setColor(axes, Color.red);
      final Position axes_p = new Position(axes);
      scene.addPosition(axes_p);

      // Add a square to the Scene.
      final Model square = new Square("Example_01");
      final Position square_p = new Position(square);
      scene.addPosition(square_p);

      // Rotate the square around its corner at the origin.
      for (int i = 0; i < 72; ++i)
      {
         // Replace the Position's matrix with a new rotation matrix.
         square_p.transform( Matrix.rotateZ(5 * i) );

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_01a_Frame%03d.ppm", i));
      }


      // Second version.
      // Reset the position's model matrix to the identity matrix.
      square_p.transform( Matrix.identity() );
      // Rotate the square around its corner at the origin.
      for (int i = 0; i < 72; ++i)
      {
         // Render again.
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_01b_Frame%03d.ppm", i));

         // Accumulate rotations on the right side of the Position's matrix.
         square_p.transform( square_p.getMatrix().times(Matrix.rotateZ(5)) );
      }


      // Third version.
      // Reset the position's model matrix to the identity matrix.
      square_p.transform( Matrix.identity() );
      // Rotate the square around its corner at the origin.
      for (int i = 0; i < 72; ++i)
      {
         // Render again.
         fb.clearFB(Color.darkGray.brighter());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_01c_Frame%03d.ppm", i));

         // Accumulate rotations on the left side of the Position's matrix.
         square_p.transform( Matrix.rotateZ(5).times(square_p.getMatrix()) );
      }
   }
}
