/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Rotate the unit square around the point (3/4, 3/4) in the
   square's model coordinate system, but with the center of
   rotation at the point (1/4, 1/4) in the world coordinate
   system.
<p>
   Also demonstrate the transformations of the coordinate system.
<p>
   See the animation <a href="../Example_08.gif">Example_08.gif</a>.
<p>
   <a href="../Example_08.gif"><img src="../Example_08.gif" alt="Example_08.gif"></a>
<p>
   For the i'th frame of the animation, the square's model matrix is
   <pre>{@code
      M = T(1/4, 1/4, 0) * Rz(5*i) * T(-3/4, -3/4, 0).
   }</pre>
*/
public class Example_08b
{
   public static void main(String[] args)
   {
      // Since this is a 2D example, we can use an orthographic camera.
      final double right  = 2.0;
      final double left   = -right;
      final double top    = right;
      final double bottom = -top;
      final Scene scene = new Scene("Example_08b",
                                 Camera.projOrtho(left, right, bottom, top));

      // Create a FrameBuffer to render our scene into.
      final int width  = 512;
      final int height = 512;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.black);

      // Create a set of x and y axes.
      final Model axes = new Axes2D(left, right, bottom, top, 8, 8);
      ModelShading.setColor(axes, Color.red);
      final Position axes_p = new Position(axes);
      scene.addPosition(axes_p);

      // Create three shorter sets of x and y axes.
      final Model sAxes1 = new Axes2D(-0.5, +0.5, -0.5, +0.5, 2, 2),
                  sAxes2 = new Axes2D(-0.5, +0.5, -0.5, +0.5, 2, 2),
                  sAxes3 = new Axes2D(-0.5, +0.5, -0.5, +0.5, 2, 2);
      ModelShading.setColor(sAxes1, new Color(150, 0, 0));
      ModelShading.setColor(sAxes2, new Color(100, 0, 0));
      ModelShading.setColor(sAxes3, new Color(100, 0, 0));
      final Position sAxes1_p = new Position(sAxes1),
                     sAxes2_p = new Position(sAxes2),
                     sAxes3_p = new Position(sAxes3);
      // Add the shorter axes to the Scene.
      scene.addPosition(sAxes1_p, sAxes2_p, sAxes3_p);

      // Add a square to the Scene.
      final Model square = new Square("Example_08b");
      final Position square_p = new Position(square);
      scene.addPosition(square_p);

      // Rotate the square around the point (3/4, 3/4) in the
      // square's model coordinate system, but with the center
      // of rotation at the point (1/4, 1/4) in the world
      // coordinate system.
      for (int i = 0; i < 72; ++i)
      {
         sAxes1_p.transform( Matrix.translate(0.25, 0.25, 0) );

         sAxes2_p.transform( Matrix.translate(0.25, 0.25, 0)
                     .times( Matrix.rotateZ(5*i) ) );

         sAxes3_p.transform( Matrix.translate(0.25, 0.25, 0)
                     .times( Matrix.rotateZ(5*i) )
                     .times( Matrix.translate(-0.75, -0.75, 0) ) );

         square_p.transform( Matrix.translate(0.25, 0.25, 0)
                     .times( Matrix.rotateZ(5*i) )
                     .times( Matrix.translate(-0.75, -0.75, 0) ) );

         sAxes1.visible = false;
         sAxes2.visible = false;
         sAxes3.visible = false;
         square.visible = false;
         fb.clearFB(Color.darkGray.darker());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_08b_Frame%03da.ppm", i));

         sAxes1.visible = true;
         fb.clearFB(Color.darkGray.darker());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_08b_Frame%03db.ppm", i));

         sAxes2.visible = true;
         fb.clearFB(Color.darkGray.darker());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_08b_Frame%03dc.ppm", i));

         sAxes3.visible = true;
         fb.clearFB(Color.darkGray.darker());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_08b_Frame%03dd.ppm", i));

         square.visible = true;
         fb.clearFB(Color.darkGray.darker());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_08b_Frame%03de.ppm", i));
      }
   }
}
