/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Rotate the letter {@link P} around the perimeter of a circle
   and keep {@link P} tangent to the circle.
<p>
   See the animation <a href="../Example_15.gif">Example_15.gif</a>.
<p>
   <a href="../Example_15.gif"><img src="../Example_15.gif" alt="Example_15.gif"></a>
<p>
   We will solve this problem two ways. Each way makes a clever use of this
   general rotation pattern,
   <pre>{@code
      Matrix m = Matrix.translate(c, d, 0)
         .times( Matrix.rotateZ(theta) )
         .times( Matrix.translate(-a, -b, 0) );
   }</pre>
   where {@code m} is the matrix that rotates a 2D model around the
   point {@code (a, b)} in the model's coordinates, and makes this
   rotation happen at the point {@code (c, d)} in world coordinates.
   <p>
*/
public class Example_15a
{
   public static void main(String[] args)
   {
      // Since this is a 2D example, we can use an orthographic camera.
      final double right  = 2.0;
      final double left   = -right;
      final double top    = right;
      final double bottom = -top;
      final Scene scene = new Scene("Example_15a",
                                 Camera.projOrtho(left, right, bottom, top));

      // Create a FrameBuffer to render our scene into.
      final int width  = 512;
      final int height = 512;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.black);

      // Create a set of x and y axes.
      final Model axes = new Axes2D(left, right, bottom, top, 8, 8);
      ModelShading.setColor(axes, Color.red);
      final Position axes_p = new Position(axes);
      scene.addPosition(axes_p);

      // Create a circle of radius 1.
      final Model circle = new Circle(1.0, 32);
      ModelShading.setColor(circle, Color.green);
      final Position circle_p = new Position(circle);
      scene.addPosition(circle_p);

      // Create a Model of the letter P.
      final Model modelP = new P();
      ModelShading.setColor(modelP, new Color(50, 150, 170));
      final Position modelP_p = new Position(modelP);
      scene.addPosition(modelP_p);


      // Rotate the letter P around the perimeter of the circle
      // and keep P tangent to the circle.
      for (int i = 0; i < 72; ++i)
      {
         // Rotate and then translate.
         modelP_p.transform( Matrix.rotateZ(6*i)
                     .times( Matrix.translate(1, 0, 0) ) );

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_15a_Frame%03d.ppm", i));
      }
   }
}
