/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Rotate the letter {@link P} around the perimeter of a circle
   and keep {@link P} tangent to the circle.
<p>
   Demonstrate the transformations of the coordinate system.
<p>
   In the last example file, we said that this code
   <pre>{@code
      modelP_p.transform( Matrix.translate(Math.cos(5*i*Math.PI/180),
                                           Math.sin(5*i*Math.PI/180),
                                           0)
                  .times( Matrix.rotateZ(5*i) ) );
   }</pre>
   when read from the top down, should be read as a sequence of
   transformations of the coordinate system. (When it is read from
   the bottom up, it should be read as a sequence of transformations
   of vertices.) This program draws a visualization of the coordinate
   systems being transformed.
<p>
   In this animation, each frame of the last animation is turned into
   four frames. The first frame represents this line of code
   <pre>{@code
      modelP_p.transform( Matrix.identity() );
   }</pre>
   which restores the coordinates system to the original one. The
   second frame represents this line of code
   <pre>{@code
      modelP_p.transform( Matrix.translate(Math.cos(5*i*Math.PI/180),
                                           Math.sin(5*i*Math.PI/180),
                                           0)
   }</pre>
   which translates the coordinate system. The third frame represents
   this line of code
   <pre>{@code
                  .times( Matrix.rotateZ(5*i) ) );
   }</pre>
   which rotates the translated coordinate system. Then the last
   frame draws the {@code P} model in the last coordinate system.
*/
public class Example_15d
{
   public static void main(String[] args)
   {
      // Since this is a 2D example, we can use an orthographic camera.
      final double right  = 2.0;
      final double left   = -right;
      final double top    = right;
      final double bottom = -top;
      final Scene scene = new Scene("Example_15d",
                                 Camera.projOrtho(left, right, bottom, top));

      // Create a FrameBuffer to render our scene into.
      final int width  = 512;
      final int height = 512;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.black);

      // Create a set of x and y axes.
      final Model axes = new Axes2D(left, right, bottom, top, 8, 8);
      ModelShading.setColor(axes, Color.red);
      final Position axes_p = new Position(axes);
      scene.addPosition(axes_p);

      // Create a circle of radius 1.
      final Model circle = new Circle(1.0, 32);
      ModelShading.setColor(circle, Color.green);
      final Position circle_p = new Position(circle);
      scene.addPosition(circle_p);

      // Create a Model of the letter P.
      final Model modelP = new P();
      ModelShading.setColor(modelP, new Color(50, 150, 170));
      final Position modelP_p = new Position(modelP);
      scene.addPosition(modelP_p);

      // Create two more sets of x and y axes with different colors.
      final Model sAxes1 = new Axes2D(-1.25, +1.25, -1.25, +1.25, 6, 6);
      final Model sAxes2 = new Axes2D(-1.25, +1.25, -1.25, +1.25, 6, 6);
      ModelShading.setColor(sAxes1, Color.white);
      ModelShading.setColor(sAxes2, Color.blue);
      final Position sAxes1_p = new Position(sAxes1);
      final Position sAxes2_p = new Position(sAxes2);
      scene.addPosition(sAxes1_p, sAxes2_p);

      // Rotate the letter P around the perimeter of the circle
      // and keep P tangent to the circle.
      for (int i = 0; i < 72; ++i)
      {
         sAxes1_p.transform( Matrix.translate(
                                     Math.cos(5*i*Math.PI/180),
                                     Math.sin(5*i*Math.PI/180),
                                     0) );

         sAxes2_p.transform( Matrix.translate(
                                     Math.cos(5*i*Math.PI/180),
                                     Math.sin(5*i*Math.PI/180),
                                     0)
                     .times( Matrix.rotateZ(5*i) ) );

         modelP_p.transform( Matrix.translate(
                                     Math.cos(5*i*Math.PI/180),
                                     Math.sin(5*i*Math.PI/180),
                                     0)
                     .times( Matrix.rotateZ(5*i) ) );

         sAxes1.visible = false;
         sAxes2.visible = false;
         modelP.visible = false;
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_15d_Frame%03da.ppm", i));

         sAxes1.visible = true;
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_15d_Frame%03db.ppm", i));

         sAxes2.visible = true;
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_15d_Frame%03dc.ppm", i));

         modelP.visible = true;
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_15d_Frame%03dd.ppm", i));
      }
   }
}
