/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Rotate the letter {@link P} around the top of the {@link P}
   while scaling it twice as large for each rotation.
<p>
   The top of {@link P} should always be at the point (0, 1).
<p>
   See the animation <a href="../Example_32a.gif">Example_32a.gif</a>.
<p>
   <a href="../Example_32a.gif"><img src="../Example_32a.gif" alt="Example_32a.gif"></a>
*/
public class Example_32
{
   public static void main(String[] args)
   {
      // Set up an orthographic camera.
      final double right  = 4.0;
      final double left   = -right;
      final double top    =  right;
      final double bottom = -top;
      final Scene scene = new Scene("Example_32",
                             Camera.projOrtho(left, right, bottom, top));

      // Create a FrameBuffer to render our scene into.
      final int width  = 512;
      final int height = 512;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.black);

      // Create a set of x and y axes.
      final Model axes = new Axes2D(left, right, bottom, top, 16, 16);
      ModelShading.setColor(axes, Color.red);
      final Position axes_p = new Position(axes);
      scene.addPosition(axes_p);

      // Create a Model of the letter P.
      final P modelP = new P();
      ModelShading.setColor(modelP, Color.blue);
      final Position modelP_p = new Position(modelP);
      scene.addPosition(modelP_p);


      // Rotate the letter P around the top of the P
      // while scaling it twice as large for each rotation.
      // The top of P should always be at the point (0, 1).
      for (int i = 0; i <= 120; i++)
      {
         final double scale = Math.pow(2, i/60.0);
         modelP_p.transform( Matrix.translate(0, 1, 0)
                     .times( Matrix.rotateZ(i*6) )
                     .times( Matrix.scale(scale, scale, 1) )
                     .times( Matrix.translate(0, -1, 0) ) );

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_32a_Frame%03d.ppm", i));
      }


      // Why does this animation act different than the one just above?
      for (int i = 0; i <= 120; ++i)
      {
         final double scale = Math.pow(2, i/60.0);
         modelP_p.transform( Matrix.translate(0, 1, 0)
                     .times( Matrix.rotateZ(i*6) )
                     .times( Matrix.translate(0, -1, 0) )
                     .times( Matrix.scale(scale, scale, 1) ) );

         // Render again.
         fb.clearFB(Color.darkGray.darker());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_32b_Frame%03d.ppm", i));
      }


      // Why is this animation the same as the first one?
      for (int i = 0; i <= 120; ++i)
      {
         final double scale = Math.pow(2, i/60.0);
         modelP_p.transform( Matrix.translate(0, 1, 0)
                     .times( Matrix.rotateZ(i*6) )
                     .times( Matrix.translate(0, -scale, 0) )
                     .times( Matrix.scale(scale, scale, 1) ) );

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_32c_Frame%03d.ppm", i));
      }


      // Try to do this animation with the transformations
      // in the following order (from top to bottom),
      // scale, translate, rotate, translate.
      for (int i = 0; i <= 120; ++i)
      {
         final double scale = Math.pow(2, i/60.0);
         // Do scale, translate, rotate, translate.






         // Render again.
         fb.clearFB(Color.darkGray.darker());
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_32d_Frame%03d.ppm", i));
      }


   }
}
