/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Use this program to experiment with two-dimensional rotations.
<p>
   Suppose you want to rotate a 2D model around the point (a, b) in
   the model's coordinates, and you want this rotation to happen at
   the point (c, d) in world coordinates. You would do that with the
   following three transformations.
   <pre>{@code
      Matrix m = Matrix.translate(c, d, 0)
         .times( Matrix.rotateZ(theta) )
         .times( Matrix.translate(-a, -b, 0) );
   }</pre>
*/
public class RotationPattern_ver2
{
   public static void main(String[] args)
   {
      // Since this is a 2D example, we can use an orthographic camera.
      final double right  = 2.0;
      final double left   = -right;
      final double top    = right;
      final double bottom = -top;
      final Scene scene = new Scene("RotationPattern",
                                 Camera.projOrtho(left, right, bottom, top));

      // Create a FrameBuffer to render our scene into.
      final int width  = 512;
      final int height = 512;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.black);

      // Create a set of x and y axes.
      final Model axes = new Axes2D(left, right, bottom, top, 8, 8);
      ModelShading.setColor(axes, Color.red);
      final Position axes_p = new Position(axes);
      scene.addPosition(axes_p);

      // Create three shorter sets of x and y axes.
      final Model sAxes1 = new Axes2D(-0.5, +0.5, -0.5, +0.5, 2, 2),
                  sAxes2 = new Axes2D(-0.5, +0.5, -0.5, +0.5, 2, 2),
                  sAxes3 = new Axes2D(-0.5, +0.5, -0.5, +0.5, 2, 2);
      ModelShading.setColor(sAxes1,Color.red);
      ModelShading.setColor(sAxes2, Color.green);
      ModelShading.setColor(sAxes3, Color.blue);
      final Position sAxes1_p = new Position(sAxes1),
                     sAxes2_p = new Position(sAxes2),
                     sAxes3_p = new Position(sAxes3);
      // Add the shorter axes to the Scene.
      scene.addPosition(sAxes1_p, sAxes2_p, sAxes3_p);

      // Add the letter F to the Scene.
      final Model fModel = new F();
      ModelShading.setColor(fModel, Color.cyan);
      final Position fModel_p = new Position(fModel);
      scene.addPosition(fModel_p);

      final Vertex vInModel = fModel.vertexList.get(7);
      final double a = vInModel.x;
      final double b = vInModel.y;
      final double c = 1.0;
      final double d = 0.0;
      final double theta = 45;

      // Illustrate the left-to-right transformation of coordinate systems.
      sAxes1_p.transform( Matrix.translate(c, d, 0) );

      sAxes2_p.transform( Matrix.translate(c, d, 0)
                  .times( Matrix.rotateZ(theta) ) );

      sAxes3_p.transform( Matrix.translate(c, d, 0)
                  .times( Matrix.rotateZ(theta) )
                  .times( Matrix.translate(-a, -b, 0) ) );

      fModel_p.transform( Matrix.translate(c, d, 0)
                  .times( Matrix.rotateZ(theta) )
                  .times( Matrix.translate(-a, -b, 0) ) );

      sAxes1.visible = false;
      sAxes2.visible = false;
      sAxes3.visible = false;
      fModel.visible = false;
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_left-to-right_a.ppm");

      sAxes1.visible = true;
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_left-to-right_b.ppm");

      sAxes2.visible = true;
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_left-to-right_c.ppm");

      sAxes3.visible = true;
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_left-to-right_d.ppm");

      fModel.visible = true;
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_left-to-right_e.ppm");


      // Illustrate the right-to-left transformation of the model's vertices.
      sAxes1.visible = false;
      sAxes2.visible = false;
      sAxes3.visible = false;
      fModel.visible = true;

      fModel_p.transform( Matrix.identity() );
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_right-to-left_a.ppm");

      fModel_p.transform( Matrix.translate(-a, -b, 0) );
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_right-to-left_b.ppm");

      fModel_p.transform( Matrix.rotateZ(theta)
                  .times( Matrix.translate(-a, -b, 0) ) );
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_right-to-left_c.ppm");

      fModel_p.transform( Matrix.translate(c, d, 0)
                  .times( Matrix.rotateZ(theta) )
                  .times( Matrix.translate(-a, -b, 0) ) );
      fb.clearFB(Color.darkGray.darker());
      Pipeline.render(scene, fb);
      fb.dumpFB2File("RotationPattern_ver2_right-to-left_d.ppm");
   }
}
